from tkinter import *  # Oberflächenelemente so zugänglich machen, dass diese ohne Zusätze auch im Hauptprogramm verwendet werden können
import threading
import time
import random
import os
import sys
if sys.platform.startswith( 'win' ):
    import winsound


class Spielfeld( Tk ):
    """Ein Objekt dieser Klasse stellt ein Spielfeld dar. VERSION: 1.1.0"""
            

    # Erstelle ein statisches Lock-Objekt zur Thread-Synchronisation (für Sound und andere Spielfeld-übergreifende Methoden)
    statischerLock = threading.RLock()

        
    def __init__( self ):
        """Konstruktor des Spielfeldes: Erzeugt ein neues Spielfeld."""
        
        # Erstelle ein separates Lock-Objekt für die Warten-Methode (falls irgendwann mehrere Spielfelder unterstützt werden sollten)
        self.lock = threading.RLock()

        # Erstelle Bedingung zum Warten auf Aktualisierung des Spielfelds
        self.bedingungSpielfeldAktualisieren = threading.Condition( self.lock )

        # Erstelle Ereignis zum Warten auf Initialisierung des Spielfelds
        self.ereignisSpielfeldInitialisieren = threading.Event()
                
        # Starte Fensterablauf
        self.fensterablauf = threading.Thread( target = self._starteFenster )
        self.fensterablauf.start()

        # Warten bis Initialisierung des Fensters fertig
        self.ereignisSpielfeldInitialisieren.wait()

 
    def _starteFenster( self ):
        """Initialisiert und startet das Fenster des Spielfelds."""
                
        # Initialisiere Fenster
        Tk.__init__( self )
    
        # Größe des Fensters festlegen
        self.breite = 800
        self.höhe = 600
        self.geometry( str( self.breite ) + "x" + str( self.höhe ) )

        # Fenster einstellen
        self.resizable( False, False )
        self.title( "Spiel" )

        # Leinwand erzeugen
        self.leinwand = Canvas( self )
        self.leinwand.place( x = 0, y = 0, width = self.breite, height = self.höhe )

        # Variable für die momentan gedrückte taste
        self.tasteGedrückt = ""

        # Variable, die angibt, ob die Pfeiltaste nach links gedrückt wird
        self.linksGedrückt = False

        # Variable, die angibt, ob die Pfeiltaste nach rechts gedrückt wird
        self.rechtsGedrückt = False

        # Variable, die angibt, ob die Pfeiltaste nach oben gedrückt wird
        self.obenGedrückt = False

        # Variable, die angibt, ob die Pfeiltaste nach unten gedrückt wird
        self.untenGedrückt = False

        # Gibt an, ob die linke Maustaste gedrückt wird
        self.linksGeklickt = False

        # Gibt an, ob die rechte Maustaste gedrückt wird
        self.rechtsGeklickt = False
	
        # Die momentane Position des Mauszeigers auf dem Spielfeld
        self.mausX = 0
        self.mausY = 0

        # Das Spielbojekt auf dem Spielfeld, auf welches der Mauszeiger momentan zeigt. "None", wenn der Mauszeiger auf kein Spielobjekt zeigt
        self.spielobjektBeiMaus = None

        # Datenstruktur für Spielobjekte und Bilder erzeugen
        self.spielobjekte = []

        # Liste für Bilder erzeugen
        self.bilder = []

        # Liste für skalierte Bilder erzeugen
        self.skalierteBilder = []

        # Liste für Leinwandbilder erzeugen
        self.leinwandbilder = []

        # Liste für Leinwandtexte erzeugen
        self.leinwandtexte = []
        
        # Der Zeitpunkt des letzten Aufrufs der Methode "warten( int millisekunden )"
        self.vorherigerZeitpunkt = 0

        # Beendenereignis festlegen
        self.protocol( "WM_DELETE_WINDOW", self.onClose )

        # Unerwünschte Tastenereignisse unterbinden (pausieren sonst das laufende Programm)
        self.bind_all("<F10>", "")
        self.bind_all("<Alt_L>", "")

        # Tastenereignisse festlegen
        self.bind_all( "<Key>", self.onKeyDown )
        self.bind_all( "<KeyRelease>", self.onKeyReleased )

        # Mausereignisse festlegen
        self.bind_all( "<Motion>", self.onMouseMoved )
        self.bind_all( "<Button-1>", self.onMouseLeftPressed )
        self.bind_all( "<Button-3>", self.onMouseRightPressed )
        self.bind_all( "<ButtonRelease-1>", self.onMouseLeftReleased )
        self.bind_all( "<ButtonRelease-3>", self.onMouseRightReleased )

        # Variable für Aktualisierung des Spielfelds
        self.spielfeldAktualisieren = False

        # Variablen für Hinzufügen eines Spielobjektes
        self.spielobjektHinzufügen = False
        self.hinzuzufügendeSpielobjekte = []
        
        # Variablen für Entfernen eines Spielobjektes
        self.spielobjektEntfernen = False
        self.zuEntfernendeSpielobjekte = []

        # Variable für alle Spielobjekte entfernen
        self.alleSpielobjekteEntfernen = False

        # Liste mit entfernten Bilder erstellen (da das Fenster automatisch neu gezeichnet wird, sobald ein Bild 
        # vollständig entfernt wird - auch wenn zwischenzeitlich die Zeichen-Methode nicht erneut aufgerufen wurde)
        self.entfernteBilder = []
                
        # Timer starten
        self.onTimer()
        
        # Benachrichtigen, dass Initialisierung des Spielfelds fertig
        self.ereignisSpielfeldInitialisieren.set()
        
        # Fensterablauf starten
        mainloop()


    def hinzufügen( self, spielobjekt ):
        """Fügt dem Spielfeld ein neues Spielobjekt hinzu."""

        # Synchronisieren
        self.bedingungSpielfeldAktualisieren.acquire()

        # Wenn Aufruf nicht aus dem Fensterablauf, warten bis Fenster-Thread Spielobjekt hinzugefügt hat
        if not threading.current_thread() is self.fensterablauf:
            self.hinzuzufügendeSpielobjekte.append( spielobjekt )
            self.spielobjektHinzufügen = True
            self.bedingungSpielfeldAktualisieren.wait()

        # ...sonst Spielobjekt unsynchronisiert hinzufügen
        else:
            self._hinzufügen( spielobjekt )
        
        # Freigeben
        self.bedingungSpielfeldAktualisieren.release()
        

    def _hinzufügen( self, spielobjekt ):
        """Fügt dem Spielfeld sofort ein neues Spielobjekt hinzu (unsynchronisiert!)."""

        # Spielobjekt hinzufügen
        self.spielobjekte.append( spielobjekt )

        # Wenn Spielobjekt ein Bild hat, Bild hinzufügen
        if (spielobjekt.bildname != None) and (spielobjekt.bildname != ""):
            self.bilder.append( { spielobjekt.bildname : PhotoImage( file = spielobjekt.bildname ) } )
        else:
            self.bilder.append( { spielobjekt.bildname : None } )
        self.skalierteBilder.append( { spielobjekt.bildname : self.bilder[ -1 ][ spielobjekt.bildname ] } )

        # Rekursiv alle untergeordneten Spielobjekte hinzufügen
        for field in spielobjekt.__dict__:
            if isinstance( spielobjekt.__dict__[ field ], Spielobjekt ):
                self._hinzufügen( spielobjekt.__dict__[ field ] )


    def entfernen( self, spielobjekt ):
        """Entfernt ein Spielobjekt vom Spielfeld."""

        # Synchronisieren
        self.bedingungSpielfeldAktualisieren.acquire()
        
        # Wenn Aufruf nicht aus dem Fensterablauf, warten bis Fenster-Thread Spielobjekt entfernt hat
        if not threading.current_thread() is self.fensterablauf:
            self.zuEntfernendeSpielobjekte.append( spielobjekt )
            self.spielobjektEntfernen = True
            self.bedingungSpielfeldAktualisieren.wait()

        # ...sonst Spielobjekt unsynchronisiert entfernen
        else:
            self._entfernen( spielobjekt )
                
        # Freigeben
        self.bedingungSpielfeldAktualisieren.release()
        

    def _entfernen( self, spielobjekt ):
        """Entfernt ein Spielobjekt vom Spielfeld. (unsynchronisiert!)"""

        # Bilder die entfernt werden merken, da sonst das Fenster automatisch neu gezeichnet wird (auch ohne Aufruf der Zeichnen-Methode)
        self.entfernteBilder.append( self.bilder[ self.spielobjekte.index( spielobjekt ) ] )
        self.entfernteBilder.append( self.skalierteBilder[ self.spielobjekte.index( spielobjekt ) ] )

        # Bilder entfernen
        self.bilder.remove( self.bilder[ self.spielobjekte.index( spielobjekt ) ] )
        self.skalierteBilder.remove( self.skalierteBilder[ self.spielobjekte.index( spielobjekt ) ] )

        # Spielobjekt entfernen
        self.spielobjekte.remove( spielobjekt )

        # Rekursiv alle untergeordneten Spielobjekte löschen
        for field in spielobjekt.__dict__:
            if isinstance( spielobjekt.__dict__[ field ], Spielobjekt ):
                self._entfernen( spielobjekt.__dict__[ field ] )
                
        
    def alleEntfernen( self ):
        """Entfernt alle Spielobjekte vom Spielfeld."""

        # Synchronisieren
        self.bedingungSpielfeldAktualisieren.acquire()

        # Wenn Aufruf nicht aus dem Fensterablauf, warten bis Fenster-Thread alle Spielobjekte entfernt hat
        if not threading.current_thread() is self.fensterablauf:
            self.alleSpielobjekteEntfernen = True
            self.bedingungSpielfeldAktualisieren.wait()

        # ...sonst Spielobjekt unsynchronisiert hinzufügen
        else:
            self._alleEntfernen()
        
        # Freigeben
        self.bedingungSpielfeldAktualisieren.release()


    def _alleEntfernen( self ):
        """Entfernt alle Spielobjekte vom Spielfeld. (unsynchronisiert!)"""

        # Alle Spielobjekte entfernen
        while len( self.spielobjekte ) > 0:
            self._entfernen( self.spielobjekte[ 0 ] )
            

    def zeichnen( self ):
        """Zeichnet das Spielfeld mit allen seinen momentan enthaltenen Spielobjekten."""

        # Synchronisieren
        self.bedingungSpielfeldAktualisieren.acquire()
        
        # Wenn Aufruf nicht aus dem Fensterablauf, warten bis Fenster-Thread Spielfeld neu gezeichnet hat
        if not threading.current_thread() is self.fensterablauf:
            self.spielfeldAktualisieren = True
            self.bedingungSpielfeldAktualisieren.wait()
        else:
            self._zeichnen()

        # Freigeben
        self.bedingungSpielfeldAktualisieren.release()


    def _zeichnen( self ):
        """Zeichnet das Spielfeld mit allen seinen momentan enthaltenen Spielobjekten sofort (unsynchronisiert!)."""

        # Bilder nachladen, wenn sie sich geändert haben und noch nicht existieren
        for i in range( len( self.spielobjekte ) ):

            if not self.spielobjekte[ i ].bildname in self.bilder[ i ]:

                if (self.spielobjekte[ i ].bildname != None) and (self.spielobjekte[ i ].bildname != ""):
                    self.bilder[ i ][ self.spielobjekte[ i ].bildname ] = PhotoImage( file = self.spielobjekte[ i ].bildname )
                else:
                    self.bilder[ i ][ self.spielobjekte[ i ].bildname ] = None
                self.skalierteBilder[ i ][ self.spielobjekte[ i ].bildname ] = self.bilder[ i ][ self.spielobjekte[ i ].bildname ]

                # DEBUG
                #print( "Bild nachgeladen!" )
        
        # Bildergröße anpassen (als Vorbereitung auf das Zeichnen)
        for i in range( len( self.spielobjekte ) ):

            # Wenn Spielobjekt ein Bild hat und sichtbar ist und sich die Bildgröße verändert hat
            if (self.bilder[ i ][ self.spielobjekte[ i ].bildname ] != None) and (not self.spielobjekte[ i ].unsichtbar) and (self._skalierteBreite( i ) != self.skalierteBilder[ i ][ self.spielobjekte[ i ].bildname ].width() or self._skalierteHöhe( i ) != self.skalierteBilder[ i ][ self.spielobjekte[ i ].bildname ].height()): 

                # DEBUG
                #print( "Bildgröße verändert: ", self.skalierteBilder[ i ][ self.spielobjekte[ i ].bildname ].width(), self.skalierteBilder[ i ][ self.spielobjekte[ i ].bildname ].height(), self._skalierteBreite( i ), self._skalierteHöhe( i ), self.spielobjekte[ i ].bildname )

                skaliertesBild = self._bildgrößeVerändern( self.bilder[ i ][ self.spielobjekte[ i ].bildname ], self.spielobjekte[ i ].breite, self.spielobjekte[ i ].höhe )
                self.skalierteBilder[ i ][ self.spielobjekte[ i ].bildname ] = skaliertesBild

        # Bilder für neuen Frame zeichnen
        self.neueLeinwandbilder = []
        self.neueLeinwandtexte = []
        for i in range( len( self.spielobjekte ) ):
            if not self.spielobjekte[ i ].unsichtbar:
                if self.bilder[ i ][ self.spielobjekte[ i ].bildname ] != None:
                    self.neueLeinwandbilder.append( self.leinwand.create_image( self.spielobjekte[ i ].positionX, self.spielobjekte[ i ].positionY, image = self.skalierteBilder[ i ][ self.spielobjekte[ i ].bildname ], anchor = NW ) )
                if (self.spielobjekte[ i ].text != None) and (self.spielobjekte[ i ].text != ""):
                    self.neueLeinwandtexte.append( self.leinwand.create_text( self.spielobjekte[ i ].positionX, self.spielobjekte[ i ].positionY, text = self.spielobjekte[ i ].text, fill = self.spielobjekte[ i ].textfarbe, anchor = SW ) )

        # Bilder von altem Frame löschen
        for leinwandbild in self.leinwandbilder:
            self.leinwand.delete( leinwandbild )
        for leinwandtext in self.leinwandtexte:
            self.leinwand.delete( leinwandtext )

        # Neuer Frame wurde gezeichnet und wird zu altem Frame
        self.leinwandbilder = self.neueLeinwandbilder
        self.leinwandtexte = self.neueLeinwandtexte

        # Entfernte Bilder entgültig löschen (hier darf das Fenster auch nochmal neu gezeichnet werden, da hier ohnehin alle
        # Zeichen-Operationen erfolgen)
        while len( self.entfernteBilder ) > 0:
            self.entfernteBilder.remove( self.entfernteBilder[ 0 ] )

        # Spielobjekt auf das der Mauszeiger zeigt merken
        self.spielobjektBeiMaus = None
        self.spielobjekte.reverse()
        for spielobjekt in self.spielobjekte:
            if (not spielobjekt.unsichtbar) and (self.mausX >= spielobjekt.positionX) and (self.mausX <= spielobjekt.positionX + spielobjekt.breite) and (self.mausY >= spielobjekt.positionY) and (self.mausY <= spielobjekt.positionY + spielobjekt.höhe):
                self.spielobjektBeiMaus = spielobjekt
                break
        self.spielobjekte.reverse()
        

    def warten( self, millisekunden ):
        """Wartet solange, bis die in Millisekunden angegebene Zeit seit dem letzten Aufruf der Methode verstichen ist."""

        # Synchronisieren
        self.lock.acquire()
        
        wartezeit = millisekunden - ((time.time() * 1000) - self.vorherigerZeitpunkt);
        if wartezeit > 0:

            # DEBUG
            #print( "Warten: ", float( wartezeit ) / 1000 )

            time.sleep( float( wartezeit ) / 1000 )

        self.vorherigerZeitpunkt = time.time() * 1000
        
        # Freigeben
        self.lock.release()
        

    @staticmethod
    def zufallszahl( von, bis ):
        """Gibt eine Zufallszahl im angegebenen Bereich zurück (jeweils einschließlich der angegebenen Grenzen)."""

        # Synchronisieren
        Spielfeld.statischerLock.acquire()

        # Zufallszahl ermitteln
        ergebnis = random.randint( von, bis )
        
        # Freigeben
        Spielfeld.statischerLock.release()
        
        return( ergebnis )


    def onClose( self ):
        os._exit( 0 )


    def onKeyDown( self, event ):

        # DEBUG
        #print( event.char, event.keysym )

        # Gedrückte Pfeiltaste verarbeiten
        if( event.keysym == "Left" ):
            self.linksGedrückt = True;
        elif( event.keysym == "Right" ):
            self.rechtsGedrückt = True;
        elif( event.keysym == "Up" ):
            self.obenGedrückt = True;
        elif( event.keysym == "Down" ):
            self.untenGedrückt = True;
        else:
            # Gedrückte Taste verarbeiten
            self.tasteGedrückt = event.char;


    def onKeyReleased( self, event ):

        # DEBUG
        #print( event.char, event.keysym )

        # Losgelassene Pfeiltaste verarbeiten
        if( event.keysym == "Left" ):
            self.linksGedrückt = False;
        elif( event.keysym == "Right" ):
            self.rechtsGedrückt = False;
        elif( event.keysym == "Up" ):
            self.obenGedrückt = False;
        elif( event.keysym == "Down" ):
            self.untenGedrückt = False;
        else:
            # Losgelassene Taste verarbeiten
            self.tasteGedrückt = "";


    def onMouseLeftPressed( self, event ):
        self.linksGeklickt = True
        
        
    def onMouseRightPressed( self, event ):
        self.rechtsGeklickt = True


    def onMouseLeftReleased( self, event ):
        self.linksGeklickt = False


    def onMouseRightReleased( self, event ):
        self.rechtsGeklickt = False


    def onMouseMoved( self, event ):
        self.mausX = event.x
        self.mausY = event.y
        

    def onTimer( self ):

        # Synchronisieren
        self.bedingungSpielfeldAktualisieren.acquire()
            
        # Wenn Spielobjekt hinzugefügt werden soll
        if self.spielobjektHinzufügen == True:

            # Spielobjekt hinzufügen
            self.spielobjektHinzufügen = False
            while len( self.hinzuzufügendeSpielobjekte ) > 0:
                self._hinzufügen( self.hinzuzufügendeSpielobjekte.pop( 0 ) )

            # Benachrichtigen, dass Aktualisierung des Spielfelds fertig
            self.bedingungSpielfeldAktualisieren.notify()
            
        # Wenn Spielobjekt entfernt werden soll
        if self.spielobjektEntfernen == True:

            # Spielobjekt entfernen
            self.spielobjektEntfernen = False
            while len( self.zuEntfernendeSpielobjekte ) > 0:
                self._entfernen( self.zuEntfernendeSpielobjekte.pop( 0 ) )

            # Benachrichtigen, dass Aktualisierung des Spielfelds fertig
            self.bedingungSpielfeldAktualisieren.notify()
            
        # Wenn Spielobjekt entfernt werden soll
        if self.alleSpielobjekteEntfernen == True:

            # Alle Spielobjekte entfernen
            self.alleSpielobjekteEntfernen = False
            self._alleEntfernen()

            # Benachrichtigen, dass Aktualisierung des Spielfelds fertig
            self.bedingungSpielfeldAktualisieren.notify()
            
        # Wenn Spielfeld neu gezeichnet werden soll
        if self.spielfeldAktualisieren == True:

            # Spielfeld neu zeichnen
            self.spielfeldAktualisieren = False
            self._zeichnen()

            # Benachrichtigen, dass Aktualisierung des Spielfelds fertig
            self.bedingungSpielfeldAktualisieren.notify()
        
        # Freigeben
        self.bedingungSpielfeldAktualisieren.release()

        self.after( 1, self.onTimer )


    @staticmethod
    def klangAbspielen( dateiname, wiederholen = False ):
        """Spielt den Klang mit dem angegebenen Dateinamen ab. Funktioniert nur unter Windows."""

        # Synchronisieren
        Spielfeld.statischerLock.acquire()
        
        if sys.platform.startswith( 'win' ):
            if wiederholen:
                winsound.PlaySound( dateiname, winsound.SND_FILENAME | winsound.SND_ASYNC | winsound.SND_NODEFAULT | winsound.SND_NOWAIT | winsound.SND_LOOP )
            else:
                winsound.PlaySound( dateiname, winsound.SND_FILENAME | winsound.SND_ASYNC | winsound.SND_NODEFAULT | winsound.SND_NOWAIT )

        # Freigeben
        Spielfeld.statischerLock.release()
        

    @staticmethod
    def klangStoppen():
        """Stoppt das Abspielen des Klangs. Funktioniert nur unter Windows."""
        
        # Synchronisieren
        Spielfeld.statischerLock.acquire()
        
        if sys.platform.startswith( 'win' ):
            winsound.PlaySound( None, 0 )

        # Freigeben
        Spielfeld.statischerLock.release()
        

    def _bildgrößeVerändern( self, bild, breite, höhe ):
        """Skaliert bild (ungefähr) auf die angegebenen Maße breite und höhe an und gibt ein neues skaliertes Bild zurück."""

        # Breite und Höhe eines Bildes müssen größer als 0 sein
        if breite <= 0:
            breite = 1
        if höhe <= 0:
            höhe = 1

        # Ursprüngliche Breite und Höhe merken
        originalBreite = bild.width()
        originalHöhe = bild.height()

        # Breite des Bildes anpassen
        if originalBreite >= breite:
            bild = bild.zoom( 10, 1 )
            xVeränderung = round( float( bild.width() ) / breite )
            bild = bild.subsample( xVeränderung, 1 )
        else:
            bild = bild.subsample( 2, 1 )
            xVeränderung = round( float( breite ) / bild.width() )
            bild = bild.zoom( xVeränderung, 1 )

        # Höhe des Bildes anpassen    
        if originalHöhe >= höhe:
            bild = bild.zoom( 1, 10 )
            yVeränderung = round( float( bild.height() ) / höhe )
            bild = bild.subsample( 1, yVeränderung )
        else:
            bild = bild.subsample( 1, 2 )
            yVeränderung = round( float( höhe ) / bild.height() )
            bild = bild.zoom( 1, yVeränderung )

        # Bild zurückgeben
        return bild


    def _skalierteBreite( self, spielobjektNr ):
        originalBreite = self.bilder[ spielobjektNr ][ self.spielobjekte[ spielobjektNr ].bildname ].width()
        breite = self.spielobjekte[ spielobjektNr ].breite
        if( breite <= 0 ):
            breite = 1
        if originalBreite >= breite:
            bildBreite = originalBreite * 10
            xVeränderung = round( float( bildBreite ) / breite )
            return( int( (bildBreite + xVeränderung - 1) / xVeränderung ) )
        else:
            bildBreite = int( (originalBreite + 1) / 2 )
            xVeränderung = round( float( breite ) / bildBreite )
            return( bildBreite * xVeränderung )

        
    def _skalierteHöhe( self, spielobjektNr ):
        originalHöhe = self.bilder[ spielobjektNr ][ self.spielobjekte[ spielobjektNr ].bildname ].height()
        höhe = self.spielobjekte[ spielobjektNr ].höhe
        if( höhe <= 0 ):
            höhe = 1
        if originalHöhe >= höhe:
            bildHöhe = originalHöhe * 10
            yVeränderung = round( float( bildHöhe ) / höhe )
            return( int( (bildHöhe + yVeränderung - 1) / yVeränderung ) )
        else:
            bildHöhe = int( (originalHöhe + 1) / 2 )
            yVeränderung = round( float( höhe ) / bildHöhe )
            return( bildHöhe * yVeränderung )


from Spielobjekt import Spielobjekt
